# Copyright (C) 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sample that implements device registration for the Google Assistant API."""

import json
import logging
import os

import click
import google.oauth2.credentials
import google.auth.transport.requests


ASSISTANT_API_VERSION = 'v1alpha2'
logging.basicConfig(format='', level=logging.INFO)


def failed_request_exception(message, r):
    """Build ClickException from a failed request."""
    try:
        resp = json.loads(r.text)
        message = '%s: %d\n%s' % (message, resp['error']['code'],
                                  resp['error']['message'])
        return click.ClickException(message)
    except ValueError:
        # fallback on raw text response if error is not structured.
        return click.ClickException('%s: %d\n%s' % (message,
                                                    r.status_code,
                                                    r.text))


def build_api_url(api_endpoint, api_version, project_id):
    return 'https://%s/%s/projects/%s' % (api_endpoint,
                                          api_version,
                                          project_id)


def build_client_from_context(ctx):
    project_id = ctx.obj['PROJECT_ID']
    api_url = build_api_url(ctx.obj['API_ENDPOINT'],
                            ctx.obj['API_VERSION'],
                            project_id)
    session = (ctx.obj['SESSION'] or
               google.auth.transport.requests.AuthorizedSession(
                   ctx.obj['CREDENTIALS']
               ))
    return session, api_url, project_id


def pretty_print_model(devicemodel):
    """Prints out a device model in the terminal by parsing dict."""
    PRETTY_PRINT_MODEL = """Device Model ID: %(deviceModelId)s
        Project ID: %(projectId)s
        Device Type: %(deviceType)s"""
    logging.info(PRETTY_PRINT_MODEL % devicemodel)
    if 'traits' in devicemodel:
        for trait in devicemodel['traits']:
            logging.info('        Trait %s' % trait)
    else:
        logging.info('No traits')
    logging.info('')  # Newline


def pretty_print_device(device):
    """Prints out a device instance in the terminal by parsing dict."""
    logging.info('Device Instance ID: %s' % device['id'])
    if 'nickname' in device:
        logging.info('    Nickname: %s' % device['nickname'])
    if 'modelId' in device:
        logging.info('    Model: %s' % device['modelId'])
    logging.info('')  # Newline


@click.group()
@click.option('--project-id', required=True,
              help='Enter the Google Developer Project ID that you want to '
              'use with the registration tool. If you don\'t use this flag, '
              'the tool will use the project listed in the '
              'client_secret_<client-id>.json file you specify with the '
              '--client-secrets flag.')
@click.option('--verbose', flag_value=True,
              help='Shows detailed JSON response')
@click.option('--api-endpoint', default='embeddedassistant.googleapis.com',
              show_default=True,
              help='Hostname for the Google Assistant API. Do not use this '
              'flag unless explicitly instructed.')
@click.option('--credentials', show_default=True,
              default=os.path.join(click.get_app_dir('google-oauthlib-tool'),
                                   'credentials.json'),
              help='File location of the generated credentials file. The '
              'google-oauthlib-tool generates this file after authorizing '
              'the user with the client_secret_<client-id>.json file. This '
              'credentials file authorizes access to the Google Assistant '
              'API. You can use this flag if the credentials were generated '
              'in a location that is different than the default.')
@click.pass_context
def cli(ctx, project_id, verbose, api_endpoint, credentials):
    try:
        with open(credentials, 'r') as f:
            c = google.oauth2.credentials.Credentials(token=None,
                                                      **json.load(f))
            http_request = google.auth.transport.requests.Request()
            c.refresh(http_request)
    except Exception as e:
        raise click.ClickException('Error loading credentials: %s.\n'
                                   'Run google-oauthlib-tool to initialize '
                                   'new OAuth 2.0 credentials.' % e)
    ctx.obj['API_ENDPOINT'] = api_endpoint
    ctx.obj['API_VERSION'] = ASSISTANT_API_VERSION
    ctx.obj['SESSION'] = None
    ctx.obj['PROJECT_ID'] = project_id
    ctx.obj['CREDENTIALS'] = c
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)


@cli.command()
@click.option('--model', required=True,
              help='Enter a globally-unique identifier for this device model; '
              'you should use your project ID as a prefix to help avoid '
              'collisions over the range of all projects (for example, '
              '"my-dev-project-my-led1").')
@click.option('--type', type=click.Choice(['LIGHT', 'SWITCH', 'OUTLET']),
              required=True,
              help='Select the type of device hardware that best aligns with '
              'your device. Select LIGHT if none of the listed choices aligns '
              'with your device.')
@click.option('--trait', multiple=True,
              help='Add traits (abilities) that the device supports. Pass '
              'this flag multiple times to create a list of traits. Refer to '
              'https://developers.google.com/assistant/sdk/reference/traits/ '
              'for a list of supported traits.')
@click.option('--manufacturer', required=True,
              help='Enter the manufacturer\'s name in this field (for '
              'example, "Assistant SDK developer"). This information may be '
              'shown in the Assistant settings and internal analytics.')
@click.option('--product-name', required=True,
              help='Enter the product name in this field (for example, '
              '"Assistant SDK light").')
@click.option('--description',
              help='Enter a description of the product in this field (for '
              'example, "Assistant SDK light device").')
@click.option('--device', required=True,
              help='Enter an identifier for the device instance. This ID must '
              'be unique within all of the devices registered under the same '
              'Google Developer project.')
@click.option('--nickname',
              help='Enter a nickname for the device. You can use this name '
              'when talking to your Assistant to refer to this device.')
@click.option('--client-type', type=click.Choice(['SERVICE', 'LIBRARY']),
              required=True,
              help='Select the type of the client. Use SERVICE if using '
              'the Google Assistant Service or LIBRARY if using '
              'the Google Assistant Library.')
@click.pass_context
def register(ctx, model, type, trait, manufacturer, product_name, description,
             device, nickname, client_type):
    """Registers a device model and instance.

    Device model fields can only contain letters, numbers, and the following
    symbols: period (.), hyphen (-), underscore (_), space ( ) and plus (+).
    The first character of a field must be a letter or number.

    Device instance fields must start with a letter or number. The device ID
    can only contain letters, numbers, and the following symbols: period (.),
    hyphen (-), underscore (_), and plus (+). The device nickname can only
    contain numbers, letters, and the space ( ) symbol.
    """
    # cache SESSION and PROJECT_ID
    # so that we don't re-create them between commands
    ctx.obj['SESSION'] = google.auth.transport.requests.AuthorizedSession(
        ctx.obj['CREDENTIALS']
    )
    ctx.invoke(register_model,
               model=model, type=type, trait=trait,
               manufacturer=manufacturer,
               product_name=product_name,
               description=description)
    ctx.invoke(register_device, device=device, model=model,
               nickname=nickname, client_type=client_type)


@cli.command('register-model')
@click.option('--model', required=True,
              help='Enter a globally-unique identifier for this device model; '
              'you should use your project ID as a prefix to help avoid '
              'collisions over the range of all projects (for example, '
              '"my-dev-project-my-led1").')
@click.option('--type', type=click.Choice(['LIGHT', 'SWITCH', 'OUTLET']),
              required=True,
              help='Select the type of device hardware that best aligns with '
              'your device. Select LIGHT if none of the listed choices aligns '
              'with your device.')
@click.option('--trait', multiple=True,
              help='Add traits (abilities) that the device supports. Pass '
              'this flag multiple times to create a list of traits. Refer to '
              'https://developers.google.com/assistant/sdk/reference/traits/ '
              'for a list of supported traits.')
@click.option('--manufacturer', required=True,
              help='Enter the manufacturer\'s name in this field (for '
              'example, "Assistant SDK developer"). This information may be '
              'shown in the Assistant settings and internal analytics.')
@click.option('--product-name', required=True,
              help='Enter the product name in this field (for example, '
              '"Assistant SDK light").')
@click.option('--description',
              help='Enter a description of the product in this field (for '
              'example, "Assistant SDK light device").')
@click.pass_context
def register_model(ctx, model, type, trait,
                   manufacturer, product_name, description):
    """Registers a device model.

    Device model fields can only contain letters, numbers, and the following
    symbols: period (.), hyphen (-), underscore (_), space ( ) and plus (+).
    The first character of a field must be a letter or number.
    """
    session, api_url, project_id = build_client_from_context(ctx)
    model_base_url = '/'.join([api_url, 'deviceModels'])
    model_url = '/'.join([model_base_url, model])
    payload = {
        'device_model_id': model,
        'project_id': project_id,
        'device_type': 'action.devices.types.' + type,
    }
    if trait:
        payload['traits'] = trait
    if manufacturer:
        payload.setdefault('manifest', {})['manufacturer'] = manufacturer
    if product_name:
        payload.setdefault('manifest', {})['productName'] = product_name
    if description:
        payload.setdefault('manifest', {})['deviceDescription'] = description
    logging.debug(json.dumps(payload))
    r = session.get(model_url)
    logging.debug(r.text)
    if r.status_code == 200:
        click.echo('Updating existing device model: %s' % model)
        r = session.put(model_url, data=json.dumps(payload))
    elif r.status_code in (400, 403, 404):
        click.echo('Creating new device model')
        r = session.post(model_base_url, data=json.dumps(payload))
    else:
        raise failed_request_exception('Failed to check existing device model',
                                       r)
    if r.status_code != 200:
        raise failed_request_exception('Failed to register model', r)
    click.echo('Model %s successfully registered' % model)


@cli.command('register-device')
@click.option('--device', required=True,
              help='Enter an identifier for a device instance. If the device '
              'ID already exists, this command will update the device '
              'instance. If it does not exist, this command will create '
              'a new device instance. This ID must be unique within all '
              'of the devices registered under the same Google Developer '
              'project.')
@click.option('--model', required=True,
              help='Enter the identifier for an existing device model. This '
              'new device instance will be associated with this device model.')
@click.option('--nickname',
              help='Enter a nickname for the device. You can use this name '
              'when talking to your Assistant to refer to this device.')
@click.option('--client-type', required=True,
              type=click.Choice(['SERVICE', 'LIBRARY']),
              help='Select the type of the client. Use SERVICE if using '
              'the Google Assistant Service or LIBRARY if using '
              'the Google Assistant Library.')
@click.pass_context
def register_device(ctx, device, model, nickname, client_type):
    """Registers a device instance under an existing device model.

    Device instance fields must start with a letter or number. The device ID
    can only contain letters, numbers, and the following symbols: period (.),
    hyphen (-), underscore (_), and plus (+). The device nickname can only
    contain numbers, letters, and the space ( ) symbol.
    """
    session, api_url, project_id = build_client_from_context(ctx)
    device_base_url = '/'.join([api_url, 'devices'])
    device_url = '/'.join([device_base_url, device])
    payload = {
        'id': device,
        'model_id': model,
    }
    if client_type:
        payload['client_type'] = 'SDK_' + client_type
    if nickname:
        payload['nickname'] = nickname

    logging.debug(json.dumps(payload))
    r = session.get(device_url)
    if r.status_code == 200:
        click.echo('Updating existing device: %s' % device)
        session.delete(device_url)
        r = session.post(device_base_url, data=json.dumps(payload))
    elif r.status_code in (400, 403, 404):
        click.echo('Creating new device')
        r = session.post(device_base_url, data=json.dumps(payload))
    else:
        raise failed_request_exception('Failed to check existing device', r)
    if r.status_code != 200:
        raise failed_request_exception('Failed to register device', r)
    click.echo('Device instance %s successfully registered' % device)
    logging.debug(r.text)


@cli.command()
@click.option('--model', 'resource', flag_value='deviceModels', required=True,
              help='Enter the identifier for an existing device model.')
@click.option('--device', 'resource', flag_value='devices', required=True,
              help='Enter the identifier for an existing device instance.')
@click.argument('id')
@click.pass_context
def get(ctx, resource, id):
    """Gets all of the information (fields) for a given device model or
    instance.
    """
    session, api_url, project_id = build_client_from_context(ctx)
    url = '/'.join([api_url, resource, id])
    r = session.get(url)
    if r.status_code != 200:
        raise failed_request_exception('Failed to get resource', r)

    response = json.loads(r.text)
    if resource == 'deviceModels':
        pretty_print_model(response)
    elif resource == 'devices':
        pretty_print_device(response)
    logging.debug(r.text)


@cli.command()
@click.option('--model', 'resource', flag_value='deviceModels', required=True,
              help='Enter the identifier for an existing device model.')
@click.option('--device', 'resource', flag_value='devices', required=True,
              help='Enter the identifier for an existing device instance.')
@click.argument('id')
@click.pass_context
def delete(ctx, resource, id):
    """Delete given device model or instance.
    """
    session, api_url, project_id = build_client_from_context(ctx)
    url = '/'.join([api_url, resource, id])
    r = session.delete(url)
    if r.status_code != 200:
        raise failed_request_exception('failed to delete resource', r)
    click.echo(r.text)


@cli.command()
@click.option('--model', 'resource', flag_value='deviceModels', required=True)
@click.option('--device', 'resource', flag_value='devices', required=True)
@click.pass_context
def list(ctx, resource):
    """Lists all of the device models and/or instances associated with the
    current Google Developer project. To change the current project, use the
    devicetool's --project-id flag.
    """
    session, api_url, project_id = build_client_from_context(ctx)
    url = '/'.join([api_url, resource])
    r = session.get(url)
    if r.status_code != 200:
        raise failed_request_exception('Failed to list resources', r)

    response = json.loads(r.text)
    logging.debug(r.text)
    if resource == 'deviceModels':
        if 'deviceModels' in response:
            for devicemodel in response['deviceModels']:
                pretty_print_model(devicemodel)
        else:
            logging.info('No device models found')
    elif resource == 'devices':
        if 'devices' in response:
            for device in response['devices']:
                pretty_print_device(device)
        else:
            logging.info('No devices found')


def main():
    cli(obj={})


if __name__ == '__main__':
    main()



import logging
import sys
import json
import google.auth.transport.requests
import click
import os

def register (api_endpoint, project_id, device_model_id, device_config = None):
    device_config = device_config or os.path.join (click.get_app_dir('googlesamples-assistant'), 'device_config.json')
    try:
        with open(device_config) as f:
            device = json.load(f)
            device_id = device['id']
            device_model_id = device['model_id']
            logging.info("Using device model %s and device id %s",
                            device_model_id,
                            device_id)
        
    except Exception as e:
        logging.warning('Device config not found: %s' % e)
        logging.info('Registering device')
        if not device_model_id:
            logging.error('Option --device-model-id required '
                            'when registering a device instance.')
            sys.exit(-1)
        if not project_id:
            logging.error('Option --project-id required '
                            'when registering a device instance.')
            sys.exit(-1)
        device_base_url = (
            'https://%s/v1alpha2/projects/%s/devices' % (api_endpoint,
                                                            project_id)
        )
        device_id = str(uuid.uuid1())
        payload = {
            'id': device_id,
            'model_id': device_model_id,
            'client_type': 'SDK_SERVICE'
        }
        session = google.auth.transport.requests.AuthorizedSession(
            credentials
        )
        r = session.post(device_base_url, data=json.dumps(payload))
        if r.status_code != 200:
            logging.error('Failed to register device: %s', r.text)
            sys.exit(-1)
        logging.info('Device registered: %s', device_id)
        pathlib.Path(os.path.dirname(device_config)).mkdir(exist_ok=True)
        with open(device_config, 'w') as f:
            json.dump(payload, f)
    
    return device_id, device_model_id
    


# pip3 install python-telegram-bot
# BotFather /setprivacy disable
# add to group

import telegram

class Telegram:
    def __init__ (self, token, chat_id = None):
        self.token = token
        self.chat_id = chat_id
        self.bot = telegram.Bot(token = token)

    def show_messages (self):
        updates = self.bot.getUpdates()
        for u in updates:
            print(u.message)

    def updates (self):
        return self.bot.getUpdates()

    def send (self, msg, chat_id = None):
        msgs = self.updates ()
        if msgs and self.chat_id is None:
            self.chat_id = msgs [-1].message.chat.id

        self.bot.sendMessage (
            chat_id or self.chat_id,
            msg
        )


if __name__ == "__main__":
    import sys, os
    import time
    from telegram.error import TimedOut

    token = os.environ.get ('TELEGRAM_TOKEN')
    if not token:
        raise SystemExit ('telegram token reqired')
    chat_id = os.environ.get ('TELEGRAM_CHAT_ID')
    commit_title = os.environ.get ("CI_COMMIT_TITLE", "")
    if commit_title.find ('--tg-silent') != -1:
        sys.exit ()

    bot = Telegram (token, chat_id)
    for i in range (7):
        try:
            bot.send (' '.join (sys.argv [1:]))
        except TimedOut:
            time.sleep (2)
            continue
        break



import sys, os
from . import pathtool
import importlib
import pkg_resources
from types import ModuleType

try: _reloader = importlib.reload
except AttributeError: _reloader = reload

def add_path (directory):
	sys.path.insert (0, directory)

def remove_path (directory):
	for i, k in enumerate (sys.path):	
		if k == directory:
			sys.path.pop (i)
			break

from pkg_resources import get_provider
                                  
def from_file (hname, path):
	if sys.version_info [:2] >= (3, 5):
		import importlib.util
		spec = importlib.util.spec_from_file_location (hname, path)
		mod = importlib.util.module_from_spec (spec)
		spec.loader.exec_module (mod)		
		sys.modules [hname] = mod		
		get_provider (mod.__name__)			
	else:	
		from importlib import machinery
		loader = machinery.SourceFileLoader (hname, path)
		mod = loader.load_module ()
	return mod

def from_pkg (resource_package, resource_path):
	template = pkg_resources.resource_string (resource_package, resource_path)			
	compiled = compile (template, '', 'exec')
	mod = ModuleType ("temp")
	exec (compiled, mod.__dict__)
	return mod
						
def importer (directory, libpath, module = None):
	add_path (directory)
	fn = not libpath.endswith (".py") and libpath + ".py" or libpath
	modpath = os.path.join (directory, fn)
	hname = fn.split (".")[0]
	p = directory.replace ("\\", "/")
	if p.find (":") !=- 1:
		p = "/" + p.replace (":", "")
	while 1:
		if "skitai.mounted." + hname in sys.modules:
			p, l = os.path.split (p)
			if not l:
				raise SystemError ('module %s already imported, use reload')
			hname = l + "." + hname
		else:
			hname = "skitai.mounted." + hname
			break
	mod = from_file (hname, modpath)
	remove_path (directory)
	return mod, mod.__file__
	
def reimporter (module, directory = None, libpath = None):
	if directory is None:
		directory, libpath = os.path.split (module.__file__)
	
	try:			
		del sys.modules [module.__name__]
	except KeyError:
		return
		
	try: 
		return importer (directory, libpath)
	except:	
		sys.modules [module.__name__] = module
		raise

#----------------------------------------------
# will be deprecated
#----------------------------------------------
	
def reloader (module):
	directory = os.path.split (module.__file__) [0]
	add_path (directory)
	_reloader (module)
	remove_path (directory)
	
def importer_old (directory, libpath):	
	sys.path.insert(0, directory)	
	__import__ (libpath, globals ())
	libpath, abspath = pathtool.modpath (libpath)
	module = sys.modules [libpath]	
	if abspath [-4:] in (".pyc", ".pyo"):
		abspath = abspath [:-1]		
	sys.path.pop (0)		
	return module, abspath



import os

class IPSECError (Exception): pass

def execute (cmd):
	class Result: pass
	
	r = Result ()
	f = os.popen (cmd)
	r.data = f.read ()
	r.exitcode = f.close ()
	if r.exitcode is not None or r.data.find ("ERR") != -1:
		raise IPSECError("[%s] %s" % (r.exitcode, r.data))
	return r

	
class IPSEC:
	def __init__ (self):
		self.collect_existing_policies ()
		self.dirty_filter = True
		self.filters = []
	
	def collect_existing_policies (self):
		self.allpoicies = {}
		try:
			r = execute ('netsh ipsec static show policy all')
		except IPSECError as why:	
			if why.errno.find ("[05072]") != -1 or why.errno.find ("No Policies") != -1: # no policies
				return
												
		for each in r.data.split ("\n\n") [:-2]:
			for line in each.strip ().split ("\n"):
				k, v = [x.strip () for x in line.split (":", 1)]
				if k == "Policy Name":
					curkey = v
					self.allpoicies [curkey] = {}
				self.allpoicies [curkey][k] = v
					
	def hasPolicy (self, name):
		return name in self.allpoicies
	
	def importPolicy (self, file):
		r = execute ('netsh ipsec static importpolicy file="%s"' % file)		
	
	def exportPolicy (self, file):
		r = execute ('netsh ipsec static exportpolicy file="%s"' % file)
				
	def addPolicy (self, name, assign = True, pollinginterval = 180, description = "N/A"):
		r = execute ('netsh ipsec static add policy name="%s" assign=%s pollinginterval=%d description="%s"' % (name, assign and "yes" or "no", pollinginterval, description))
		self.collect_existing_policies ()
	
	def setPolicy (self, name, newname, assign = True, pollinginterval = 180, description = "N/A"):
		r = execute ('netsh ipsec static set policy name="%s" newname="%s" assign=%s pollinginterval=%d description="%s"' % (name, newname, assign and "yes" or "no", pollinginterval, description))
		self.collect_existing_policies ()
			
	def deletePolicy (self, name):
		r = execute ('netsh ipsec static delete policy name="%s"' % (name,))
		self.collect_existing_policies ()
		
	def getPolicy (self, name):
		return self.allpoicies.get (name, None)
		
	def addFilterAction (self, name, action, description = "N/A"):
		if action not in ("permit", "block", "negotiate"):
			raise ValueError("Must be one of permit, block, negotiate")
		try: 
			r = execute ('netsh ipsec static add filteraction name="%s" action=%s description="%s"' % (name, action, description))
		except Exception as why:
			if str (why).find ("05014") != -1:
				pass
			else:
				raise
	
	def setFilterAction (self, name, newname, action, description = "N/A"):
		if action not in ("permit", "block", "negotiate"):
			raise ValueError("Must be one of permit, block, negotiate")
		r = execute ('netsh ipsec static add filteraction name="%s" newname="%s" action=%s description="%s"' % (name, newname, action, description))
		
	def deleteFilterAction (self, name):
		r = execute ('netsh ipsec static delete filteraction name="%s"' % (name,))
	
	def hasFilterAction (self, name):
		try:
			r = execute ('netsh ipsec static show filteraction name="%s"' % (name,))
		except IPSECError:
			return False
		return True		
		
	def addFilterList (self, name, description = "N/A"):
		try:	
			r = execute ('netsh ipsec static add filterlist name="%s" description="%s"' % (name, description))		
		except Exception as why:
			if str (why).find ("05010") != -1:
				pass
			else:
				raise
	
	def hasFilterList (self, name, description = "N/A"):
		try:
			r = execute ('netsh ipsec static show filterlist name="%s"' % (name,))
		except IPSECError:
			return False
		return True
		
	def setFilterList (self, name, newname, description = "N/A"):
		r = execute ('netsh ipsec static add filterlist name="%s" newname="%s" description="%s"' % (name, newname, description))
		
	def deleteFilterList (self, name):
		r = execute ('netsh ipsec static delete filterlist name="%s"' % name)
	
	def getFilters (self, flName):
		if not self.dirty_filter: 
			return self.filters
			
		r = execute ('netsh ipsec static show filterlist name="%s" level=verbose' % flName)
		s = r.data.find ("---------")
		if s == -1:
			return False
		
		data = r.data [s + 10:]
		filters = []
		
		srcMask = "255.255.255.255"
		dstMask = "255.255.255.255"
		for each in data.split ("\n\n") [:-1]:
			for line in each.strip ().split ("\n"):				
				k, v = [x.strip () for x in line.split (":", 1)]				
				if k == "Source IP Address":
					if v == "<My IP Address>": v = "Me"
					srcAddr = v						
				elif k == "Source Mask":	
					srcMask = v
				elif k == "Destination IP Address":	
					if v == "<My IP Address>": v = "Me"
					dstAddr = v
				elif k == "Destination Mask":	
					dstMask = v
				elif k == "Protocol":	
					protocol = v					
				elif k == "Source Port":	
					if v == "ANY": v = "0"
					srcPort = v					
				elif k == "Destination Port":					
					if v == "ANY": v = "0"
					dstPort = v
			filters.append ((srcAddr, dstAddr, protocol, srcPort, dstPort, srcMask, dstMask))
		
		self.filters = filters
		self.dirty_filter = False
		return filters
			
	def hasFilter (self, flName, srcAddr = "1.1.1.1", dstAddr="Me", protocol="TCP", srcPort="0", dstPort="80", srcMask="255.255.255.255", dstMask="255.255.255.255"):
		for each in self.getFilters (flName):
			if each == (srcAddr, dstAddr, protocol, srcPort, dstPort, srcMask, dstMask):
				return True
		return False		
		
	def addFilter (self, flName, srcAddr = "1.1.1.1", dstAddr="Me", protocol="TCP", srcPort="0", dstPort="80", srcMask="255.255.255.255", dstMask="255.255.255.255", description = "N/A"):
		r = execute (
			'netsh ipsec static add filter filterlist="%s" srcaddr=%s srcmask=%s dstaddr=%s dstmask=%s protocol=%s srcport=%s dstport=%s description="%s"' % (
				flName, srcAddr, srcMask, dstAddr, dstMask, protocol, srcPort, dstPort, description
			)
		)
		self.dirty_filter = True
	
	def deleteFilter (self, flName, srcAddr = "1.1.1.1", dstAddr="Me", protocol="TCP", srcPort="0", dstPort="80", srcMask="255.255.255.255", dstMask="255.255.255.255"):
		r = execute (
			'netsh ipsec static delete filter filterlist="%s" srcaddr=%s srcmask=%s dstaddr=%s dstmask=%s protocol=%s srcport=%s dstport=%s' % (
				flName, srcAddr, srcMask, dstAddr, dstMask, protocol, srcPort, dstPort
			)
		)
		self.dirty_filter = True
	
	def addRule (self, name, policy, flName, faName, activate = True, description = "N/A"):
		r = execute ('netsh ipsec static add rule name="%s" policy="%s" filterlist="%s" filteraction="%s" activate=%s description="%s"' % (name, policy, flName, faName, activate and "yes" or "no", description))
	
	def setRule (self, name, newname, policy, flName, faName, activate = True, description = "N/A"):
		r = execute ('netsh ipsec static add rule name="%s" newname="%s" policy="%s" filterlist="%s" filteraction="%s" activate=%s description="%s"' % (name, newname, policy, flName, faName, activate and "yes" or "no", description))
		
	def deleteRule (self, name, policy):
		r = execute ('netsh ipsec static delete rule name="%s" policy="%s""' % (name, policy))	
	
	def hasRule (self, name, policy):
		try:
			r = execute ('netsh ipsec static show rule name="%s" policy="%s""' % (name, policy))
		except IPSECError:
			return False
		return True
		


#!/usr/bin/env python3
import houndify
import sys



CLIENT_ID = sys.argv[1]
CLIENT_KEY = sys.argv[2]
BUFFER_SIZE = 512

#
# Simplest HoundListener; just print out what we receive.
# You can use these callbacks to interact with your UI.
#
class MyListener(houndify.HoundListener):
  def onPartialTranscript(self, transcript):
    print("Partial transcript: " + transcript)
  def onFinalResponse(self, response):
    print("Final response: " + str(response))
  def onError(self, err):
    print("Error: " + str(err))


client = houndify.StreamingHoundClient(CLIENT_ID, CLIENT_KEY, "test_user")
client.setLocation(37.388309, -121.973968)

## Uncomment the lines below to see an example of using a custom
## grammar for matching.  Use the file 'turnthelightson.wav' to try it.
# clientMatches = [ {
#   "Expression" : '([1/100 ("can"|"could"|"will"|"would")."you"].[1/10 "please"].("turn"|"switch"|(1/100 "flip"))."on".["the"].("light"|"lights").[1/20 "for"."me"].[1/20 "please"])|([1/100 ("can"|"could"|"will"|"would")."you"].[1/10 "please"].[100 ("turn"|"switch"|(1/100 "flip"))].["the"].("light"|"lights")."on".[1/20 "for"."me"].[1/20 "please"])|((("i".("want"|"like"))|((("i".["would"])|("i\'d")).("like"|"want"))).["the"].("light"|"lights").["turned"|"switched"|("to"."go")|(1/100"flipped")]."on".[1/20"please"])"',
#   "Result" : { "Intent" : "TURN_LIGHT_ON" },
#   "SpokenResponse" : "Ok, I\'m turning the lights on.",
#   "SpokenResponseLong" : "Ok, I\'m turning the lights on.",
#   "WrittenResponse" : "Ok, I\'m turning the lights on.",
#   "WrittenResponseLong" : "Ok, I\'m turning the lights on."
# } ]
# client.setHoundRequestInfo('ClientMatches', clientMatches)

client.start(MyListener())

while True:
  samples = sys.stdin.buffer.read(BUFFER_SIZE)
  if len(samples) == 0: break
  if client.fill(samples): break
  
client.finish()


import os
import signal, psutil

def thread (thread):
    import ctypes    
    if not thread.isAlive():
        return
    exc = ctypes.py_object(SystemExit)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(thread.ident), exc)
    if res == 0:
        # raise ValueError ("nonexistent thread id")
        pass
    elif res > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, None)
        raise SystemError ("PyThreadState_SetAsyncExc failed")

def process (p, sig = signal.SIGTERM):
    if isinstance (p, int):
        # resolve to process
        try:
            p = psutil.Process (p)        
        except psutil.NoSuchProcess:
            return    
    p.send_signal (sig)

# utils -------------------------------------------
def child_processes (parent_pid = os.getpid (), sig = signal.SIGTERM):    
    try:
        parent = psutil.Process (parent_pid)
    except psutil.NoSuchProcess:
        return
    children = parent.children (recursive = True)
    for process in children:
        process.send_signal (sig)


from django.contrib import admin
from django.utils.safestring import mark_safe
from django.forms import Widget
from django.contrib.admin import SimpleListFilter
from django.db.models import Q, F, Count
import mimetypes

def set_title (title):
    admin.site.site_title = title
    admin.site.site_header = "{}".format (title)
    admin.site.index_title = "{} Management Console".format (title)

# widgets -----------------------------------------------------
def get_type (path):
    return mimetypes.guess_type (os.path.basename (path))[0]

def ImageWidget (width = 360):
    class _ImageWidget(Widget):
        def render(self, name, value, **noneed):
            return value and mark_safe ('<img src="{}" width="{}">'.format (value, width)) or 'No Image'
    return _ImageWidget

class LinkWidget(Widget):
    def render(self, name, value, **noneed):
        return value and mark_safe ('<a href="{}">{}</a> [<a href="{}" target="_blank">새창</a>]'.format (value, value, value)) or 'No Image'

def VideoWidget (video_width = 320, video_height = 240):
    class _VideoWidget(Widget):
        def render(self, name, value, **noneed):
            return value and mark_safe (
                '<video width="{}" height="{}" controls><source src="{}" type="{}"></video>'.format (
                    video_width, video_height, value, get_type (value)
                )
            ) or 'No Video'
    return _VideoWidget

class AudioWidget(Widget):
    def render(self, name, value, **noneed):
        return value and mark_safe ('<audio controls><source src="{}" type="{}"></audio>'.format (
            value, get_type (value)
            )
        ) or 'No Audio'


# filter prototypes -------------------------------------------
class CountFilter(SimpleListFilter):
    title = None
    parameter_name = None
    _countable_realted_name = None
    _filter = {}
    _options = [1,3,5,10,20,30]

    def create_action_count_filter (self):
        return [(i, '{}회 이상'.format (i)) for i in self._options]

    def lookups (self, request, model_admin):
        return self.create_action_count_filter ()

    def queryset (self, request, queryset):
        value = self.value()
        if value:
            return queryset.select_related ().annotate (buy_count = Count (self._countable_realted_name, **self._filter)).filter (buy_count__gt = value)
        return queryset


class NullFilter (SimpleListFilter):
    title = None
    parameter_name = None
    _field_name = None
    _options = {'NULL': 'Null', 'NOTNULL': 'Not Null'}

    def lookups (self, request, model_admin):
        return [('t', self._options ['NULL']), ('f', self._options ['NOTNULL'])]

    def queryset (self, request, queryset):
        value = self.value()
        if value:
            return queryset.filter (**{'{}__isnull'.format (self._field_name): value == 't' and True or False})
        return queryset


class StackedInline (admin.StackedInline):
    can_delete = False
    show_change_link = True


# ModelAdmin -------------------------------------------
class ModelAdmin (admin.ModelAdmin):
    image_width = 320
    video_width = 320
    enable_add = True
    enable_delete = True
    enable_change = True

    list_per_page = 100
    list_max_show_all = 200

    def has_add_permission(self, request, obj=None):
        return self.enable_add

    def has_delete_permission(self, request, obj=None):
        return self.enable_delete

    def has_change_permission(self, request, obj=None):
        return self.enable_change

    def before_changelist_view (self, queryset, context):
        pass

    def changelist_view (self, request, extra_context = None):
        r = super ().changelist_view (request)
        if hasattr (r, 'context_data') and 'cl' in r.context_data:
            self.before_changelist_view (r.context_data ['cl'].queryset, r.context_data)
        return r

    def save_model (self, request, obj, form, change):
        return super().save_model(request, obj, form, change)

    def formfield_for_dbfield (self, db_field, request, **kwargs):
        if 'widget' not in kwargs:
            if db_field.name.endswith ('image'):
                kwargs ['widget'] = ImageWidget (self.image_width)
                return db_field.formfield(**kwargs)
            elif db_field.name.endswith ('video'):
                kwargs ['widget'] = VideoWidget (self.video_width)
                return db_field.formfield(**kwargs)
            elif db_field.name.endswith ('audio'):
                kwargs ['widget'] = AudioWidget
                return db_field.formfield(**kwargs)
            elif db_field.name.endswith ('url'):
                kwargs ['widget'] = LinkWidget
                return db_field.formfield(**kwargs)
        return super ().formfield_for_dbfield (db_field, request, **kwargs)


from base64 import b64decode, b64encode
from hmac import new as hmac
import json
import hashlib
try:
	import ujson as json
except ImportError:
	import json

def add_padding (s):
	paddings = 4 - (len (s) % 4)
	if paddings != 4:
		s += "=" * paddings
	return s

def get_claim (secret_key, token):
	header, claim, sig = token.split (".")
	jheader = json.loads (b64decode (add_padding (header)).decode ("utf8"))
	alg = jheader.get ("alg")
	if not alg or alg [:2] != "HS":
		raise TypeError ("Unknown Algorithm")
	hash_method = getattr (hashlib, "sha" + alg [2:])
	mac = hmac (secret_key, None, hash_method)
	mac.update (("%s.%s" % (header, claim)).encode ("utf8"))
	if mac.digest() != b64decode (add_padding (sig)):
		raise ValueError ("Verificationn Failed")
	return json.loads (b64decode (add_padding (claim)).decode ("utf8"))

def gen_token (secret_key, claim, alg = "HS256"):
	header = b64encode (json.dumps ({"alg": alg, "typ": "JWT"}).encode ("utf8")).rstrip (b'=')
	claim = b64encode (json.dumps (claim).encode ("utf8")).rstrip (b'=')
	hash_method = getattr (hashlib, "sha" + alg [2:])
	mac = hmac (secret_key, None, hash_method)
	mac.update (header + b"." + claim)
	sig = b64encode (mac.digest()).rstrip (b'=')
	return (header + b"." + claim + b"." + sig).decode ("utf8")


if __name__ == "__main__":
	sk = b"8fa06210-e109-11e6-934f-001b216d6e71"
	token = gen_token (sk, {'user': 'Hans Roh', 'roles': ['user']}, "HS256")

	# eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJyb2xlcyI6IFsidXNlciJdLCAidXNlciI6ICJIYW5zIFJvaCJ9.ls7R30APuI9QkeI0nVa0YOV3fBo2SlFg0ESzdyPzuv0
	print (token)
	print (get_claim (sk, token))

	"""
	import requests
	f = requests.get (
		"http://127.0.0.1:5000/lufex",
		headers ={"Authorization": "Bearer %s" % token}
	)
	print (f)
	"""


##############################################################################
# Copyright 2019 SoundHound, Incorporated.  All rights reserved.
##############################################################################
import base64
import hashlib
import hmac
import http.client
import json
import threading
import time
import uuid
import urllib.parse
import struct
import gzip


try:
  import pySHSpeex
except ImportError:
  pass

HOUND_SERVER = "api.houndify.com"
TEXT_ENDPOINT = "/v1/text"
VOICE_ENDPOINT = "/v1/audio"
VERSION = "1.2.5"


class _BaseHoundClient(object):

    def __init__(self, clientID, clientKey, userID, hostname, proxyHost, proxyPort, proxyHeaders):
      self.clientID = clientID
      self.clientKey = base64.urlsafe_b64decode(clientKey)
      self.userID = userID
      self.hostname = hostname
      self.proxyHost = proxyHost
      self.proxyPort = proxyPort
      self.proxyHeaders = proxyHeaders
      self.gzip = True
      
      self.HoundRequestInfo = {
        "ClientID": clientID,
        "UserID": userID,
        "SDK": "python3",
        "SDKVersion": VERSION
      }


    def setHoundRequestInfo(self, key, value):
      """
      There are various fields in the HoundRequestInfo object that can
      be set to help the server provide the best experience for the client.
      Refer to the Houndify documentation to see what fields are available
      and set them through this method before starting a request
      """
      self.HoundRequestInfo[key] = value


    def removeHoundRequestInfo(self, key):
      """
      Remove request info field through this method before starting a request
      """
      self.HoundRequestInfo.pop(key, None)


    def setLocation(self, latitude, longitude):
      """
      Many domains make use of the client location information to provide
      relevant results.  This method can be called to provide this information
      to the server before starting the request.

      latitude and longitude are floats (not string)
      """
      self.HoundRequestInfo["Latitude"] = latitude
      self.HoundRequestInfo["Longitude"] = longitude
      self.HoundRequestInfo["PositionTime"] = int(time.time())


    def setConversationState(self, conversation_state):
      self.HoundRequestInfo["ConversationState"] = conversation_state
      if "ConversationStateTime" in conversation_state:
        self.HoundRequestInfo["ConversationStateTime"] = conversation_state["ConversationStateTime"]


    def _generateHeaders(self, requestInfo):
      requestID = str(uuid.uuid4())
      if "RequestID" in requestInfo:
        requestID = requestInfo["RequestID"]

      timestamp = str(int(time.time()))
      if "TimeStamp" in requestInfo:
        timestamp = str(requestInfo["TimeStamp"])

      HoundRequestAuth = self.userID + ";" + requestID
      h = hmac.new(self.clientKey, (HoundRequestAuth + timestamp).encode("utf-8"), hashlib.sha256)
      signature = base64.urlsafe_b64encode(h.digest()).decode("utf-8")
      HoundClientAuth = self.clientID + ";" + timestamp + ";" + signature

      headers = { 
        "Hound-Request-Info": json.dumps(requestInfo),
        "Hound-Request-Authentication": HoundRequestAuth,
        "Hound-Client-Authentication": HoundClientAuth
      }

      if "InputLanguageEnglishName" in requestInfo:
          headers["Hound-Input-Language-English-Name"] = requestInfo["InputLanguageEnglishName"]
      if "InputLanguageIETFTag" in requestInfo:
          headers["Hound-Input-Language-IETF-Tag"] = requestInfo["InputLanguageIETFTag"]

      return headers



class TextHoundClient(_BaseHoundClient):
    """
    TextHoundClient is used for making text queries for Hound
    """
    def __init__(self, clientID, clientKey, userID, requestInfo = dict(), hostname = HOUND_SERVER, proxyHost = None, proxyPort = None, proxyHeaders = None):
      _BaseHoundClient.__init__(self, clientID, clientKey, userID, hostname, proxyHost, proxyPort, proxyHeaders)
      self.HoundRequestInfo.update(requestInfo)
    

    def query(self, query):
      """
      Make a text query to Hound.

      query is the string of the query
      """
      headers = self._generateHeaders(self.HoundRequestInfo)
      if self.gzip:
        headers["Hound-Response-Accept-Encoding"] = "gzip";

      if self.proxyHost:
        conn = http.client.HTTPSConnection(self.proxyHost, self.proxyPort)
        conn.set_tunnel(self.hostname, headers = self.proxyHeaders) 
      else:
        conn = http.client.HTTPSConnection(self.hostname)
      
      conn.request("GET", TEXT_ENDPOINT + "?query=" + urllib.parse.quote(query), headers = headers)
      resp = conn.getresponse()

      raw_response = resp.read()

      try:
        if self.gzip:
          raw_response = gzip.decompress(raw_response)
        raw_response = raw_response.decode("utf-8")
        return json.loads(raw_response)
      except:
        return { "Error": raw_response }



class HoundListener(object):
    """
    HoundListener is an abstract base class that defines the callbacks
    that can be received while streaming speech to the server
    """
    def onPartialTranscript(self, transcript):
      """
      onPartialTranscript is fired when the server has sent a partial transcript
      in live transcription mode.  "transcript" is a string with the partial transcript
      """
      pass
    def onFinalResponse(self, response):
      """
      onFinalResponse is fired when the server has completed processing the query
      and has a response.  "response" is the JSON object (as a Python dict) which
      the server sends back.
      """
      pass
    def onError(self, err):
      """
      onError is fired if there is an error interacting with the server.  It contains
      the parsed JSON from the server.
      """
      pass



class StreamingHoundClient(_BaseHoundClient):
    """
    StreamingHoundClient is used to send streaming audio to the Hound
    server and receive live transcriptions back
    """
    def __init__(self, clientID, clientKey, userID, requestInfo = dict(), hostname = HOUND_SERVER, sampleRate = 16000, enableVAD = True, useSpeex = False, proxyHost = None, proxyPort = None, proxyHeaders = None):
      """
      clientID and clientKey are "Client ID" and "Client Key" 
      from the Houndify.com web site.
      """
      _BaseHoundClient.__init__(self, clientID, clientKey, userID, hostname, proxyHost, proxyPort, proxyHeaders)
      
      self.sampleRate = sampleRate
      self.useSpeex = useSpeex
      self.enableVAD = enableVAD
      
      self.HoundRequestInfo["PartialTranscriptsDesired"] = True
      self.HoundRequestInfo.update(requestInfo)


    def setSampleRate(self, sampleRate):
      """
      Override the default sample rate of 16 khz for audio.

      NOTE that only 8 khz and 16 khz are supported
      """
      if sampleRate == 8000 or sampleRate == 16000:
        self.sampleRate = sampleRate
      else:
        raise Exception("Unsupported sample rate")


    def start(self, listener=HoundListener()):
      """
      This method is used to make the actual connection to the server and prepare
      for audio streaming.

      listener is a HoundListener (or derived class) object
      """
      self.audioFinished = False
      self.lastResult = None
      self.buffer = bytes()

      if self.proxyHost:
        self.conn = http.client.HTTPSConnection(self.proxyHost, self.proxyPort)
        self.conn.set_tunnel(self.hostname, headers = self.proxyHeaders) 
      else:
        self.conn = http.client.HTTPSConnection(self.hostname)

      self.conn.putrequest("POST", VOICE_ENDPOINT)

      headers = self._generateHeaders(self.HoundRequestInfo)
      headers["Transfer-Encoding"] = "chunked";
      if self.gzip:
        headers["Hound-Response-Accept-Encoding"] = "gzip";

      for header in headers:
        self.conn.putheader(header, headers[header])
      self.conn.endheaders()
      
      self.callbackTID = threading.Thread(target = self._callback, args = (listener,))
      self.callbackTID.start()

      audio_header = self._wavHeader(self.sampleRate)
      if self.useSpeex:
        audio_header = pySHSpeex.Init(self.sampleRate == 8000)
      self._send(audio_header)

      
    def fill(self, data):
      """
      After successfully connecting to the server with start(), pump PCM samples
      through this method.

      data is 16-bit, 8 KHz/16 KHz little-endian PCM samples.
      Returns True if the server detected the end of audio and is processing the data
      or False if the server is still accepting audio
      """
      # buffer gets flushed on next call to start()
      if self.audioFinished and self.enableVAD:
        return True

      self.buffer += data

      # 20ms 16-bit audio frame = (2 * 0.02 * sampleRate) bytes
      frame_size = int(2 * 0.02 * self.sampleRate)
      while len(self.buffer) >= frame_size:
        frame = self.buffer[:frame_size]
        self.buffer = self.buffer[frame_size:]
        if self.useSpeex:
          frame = pySHSpeex.EncodeFrame(frame)
        
        self._send(frame)   

      return False


    def finish(self):
      """
      Once fill returns True, call finish() to finalize the transaction.  finish will
      wait for all the data to be received from the server.

      After finish() is called, you can start another request with start() but each
      start() call should have a corresponding finish() to wait for the threads
      """

      if len(self.buffer) > 0:
        frame = self.buffer
        if self.useSpeex:
          padding_size = int(2 * 0.02 * self.sampleRate) - len(self.buffer)
          frame = frame + b'\x00' * padding_size
          frame = pySHSpeex.EncodeFrame(frame)
          
        self._send(frame)

      self._send("")
      self.callbackTID.join()
      return self.lastResult


    def _callback(self, listener):
      headers = ""
      body = ""
      is_chunked = False
      chunk_size = None
      content_length = None
      transcripts_done = False
      headers_done = False

      gen = self._readline(self.conn.sock)
      while True:
        try:
          line = gen.send(chunk_size)
        except:
          break

        if self.gzip and line[:3] == b"\x1f\x8b\x08":
          line = gzip.decompress(line)

        line = line.decode("utf-8")
        
        if not headers_done:
          headers += line + "\r\n"
          header = line.strip().lower()
          if header == "transfer-encoding: chunked":
            is_chunked = True
          if "content-length" in header:
            content_length = int(header.split(" ")[1])
          if headers.endswith("\r\n\r\n"):
            headers_done = True
            chunk_size = content_length
          continue

        body += line

        if is_chunked and chunk_size is None:
          chunk_size = int(line, 16)
          continue

        chunk_size = None

        try:
          parsedMsg = json.loads(line)
        except:
          break

        if "Status" in parsedMsg and parsedMsg["Status"] == "Error":
          self.lastResult = parsedMsg
          listener.onError(parsedMsg)  
          self.audioFinished = True
          return     

        if "Format" in parsedMsg:
          if parsedMsg["Format"] == "SoundHoundVoiceSearchParialTranscript" or parsedMsg["Format"] == "HoundVoiceQueryPartialTranscript":
            listener.onPartialTranscript(parsedMsg["PartialTranscript"])
            if "SafeToStopAudio" in parsedMsg and parsedMsg["SafeToStopAudio"]:
              self.audioFinished = True
            if "Done" in parsedMsg and parsedMsg["Done"]:
              transcripts_done = True

          if parsedMsg["Format"] == "SoundHoundVoiceSearchResult" or parsedMsg["Format"] == "HoundQueryResult":
            self.lastResult = parsedMsg
            listener.onFinalResponse(parsedMsg)
            return

      self.lastResult = { "Error": body }
      listener.onError({ "Error": body })
      self.audioFinished = True


    def _wavHeader(self, sampleRate=16000):
      genHeader = "RIFF".encode("UTF-8")
      genHeader += struct.pack("<L", 36) #ChunkSize - dummy   
      genHeader += "WAVE".encode("UTF-8")
      genHeader += "fmt ".encode("UTF-8")             
      genHeader += struct.pack("<L", 16) #Subchunk1Size
      genHeader += struct.pack("<H", 1)  #AudioFormat - PCM
      genHeader += struct.pack("<H", 1)  #NumChannels
      genHeader += struct.pack("<L", sampleRate) #SampleRate
      genHeader += struct.pack("<L", 8 * sampleRate) #ByteRate
      genHeader += struct.pack("<H", 2) #BlockAlign
      genHeader += struct.pack("<H", 16) #BitsPerSample
      genHeader += "data".encode("UTF-8")
      genHeader += struct.pack("<L", 0) #Subchunk2Size - dummy

      return genHeader


    def _send(self, msg):
      if self.conn:
        if (isinstance(msg, str)): 
          msg = msg.encode("utf-8")
        chunk_size = "%x\r\n" % len(msg)
        try:
          self.conn.send(chunk_size.encode("utf-8"))
          self.conn.send(msg + "\r\n".encode("utf-8"))
        except:
          self.conn.close()
          self.conn = None


    def _readline(self, socket):
      response_buffer = bytearray()
      chunk_size = None
      separator = "\r\n".encode("utf-8")
      msg_size = 4096

      while True:
        msg = socket.recv(msg_size)
        if not msg: break
        
        response_buffer += msg
 
        while True:
          if chunk_size is not None:
            if len(response_buffer) < (chunk_size+2): break
            chunk = response_buffer[:chunk_size]
            response_buffer = response_buffer[chunk_size+2:]

          else:
            split_buffer = response_buffer.split(separator, 1)
            if len(split_buffer) == 1: break
            chunk = split_buffer[0]
            response_buffer = split_buffer[1]

          chunk_size = yield chunk

      if response_buffer: yield response_buffer

      



# pytest framework ---------------------------------------------
import requests
from rs4 import siesta, attrdict
import time
import sys
import os
import xmlrpc.client
from io import IOBase
import json
from . import apidoc
from urllib.parse import urlparse, quote
from aquests.protocols.http2.hyper import HTTPConnection

has_http3 = False
if os.name != 'nt' and sys.version_info >= (3, 6):
    try:
        from aquests.protocols.http3 import client as h3client
        from aquests.protocols.http3 import requests as h3requests
    except ImportError:
        pass
    else:
        has_http3 = True

class HTTP2Response:
    def __init__ (self, r):
        self.headers = self._rebuild_headers (r.headers)
        self.events = r.events
        self.status_code = r.status
        self.reason =  r.reason
        self.content = r.read ()

    def _rebuild_headers (self, headers):
        headers_ = attrdict.CaseInsensitiveDict ()
        for k, v in headers.items ():
            headers_ [k.decode ()] = v.decode ()
        return headers_

    @property
    def text (self):
        return self.content.decode ()

    def json (self):
        json.loads (self.text)

    def get_pushes (self):
        self.conn.get_pushes ()


class HTTP2:
    def __init__ (self, endpoint):
        self.endpoint = endpoint
        parts = urlparse (self.endpoint)
        self.conn = HTTPConnection(parts.netloc, enable_push = True, secure=parts.scheme == 'https')

    def urlencode (self, params, to_bytes = True):
        fm = []
        for k, v in list(params.items ()):
            fm.append ("%s=%s" % (quote (k), quote (str (v))))
        if to_bytes:
            return "&".join (fm).encode ("utf8")
        return "&".join (fm)

    def _rebuild_header (self, headers_, data):
        headers_ = headers_ or {}
        headers = attrdict.CaseInsensitiveDict ()
        for k, v in headers_.items ():
            headers [k] = v
        if data and headers.get ('content-type') is None:
            headers ['Content-Type'] = 'application/x-www-form-urlencoded'
        return headers

    def _request (self, method, url, data = None, headers = None):
        headers = self._rebuild_header (headers, data)
        if data:
            data = self.urlencode (data)
        self.conn.request (method.upper (), url, data, headers)
        return HTTP2Response (self.conn.get_response())

    def get (self, url, headers = {}):
        return self._request ('GET', url, headers = headers)

    def post (self, url, data, headers = {}):
        return self._request ('POST', url, data, headers)


class HTTP3Response (HTTP2Response):
    def __init__ (self, r):
        self.headers = self._rebuild_headers (r.headers)
        try:
            self.status_code = int (self.headers.get (':status'))
        except TypeError:
            raise h3client.ConnectionClosed
        self.reason = ''
        self.content = r.data
        self.events = r.events
        self._promises = r.promises

    def get_pushes (self):
        class Promise:
            def __init__ (self, headers):
                self.path = headers.get (':path').encode ()
        return [Promise (headers) for headers in self._promises]


class HTTP3 (HTTP2):
    def __init__ (self, endpoint):
        self.endpoint = endpoint
        parts = urlparse (self.endpoint)
        self.conn = h3client.Connection (parts.netloc)

    def _request (self, method, url, data = None, headers = None):
        headers = self._rebuild_header (headers, data)
        if data:
            data = self.urlencode (data)
        resp = self.conn.request (method.upper (), url, data, headers)
        return HTTP3Response (resp)

    def MultiCall (self):
        return h3requests.MultiCall (self.endpoint)


class Target:
    # f = Target ('http://localhost:')
    # f.get ("/v1/accounts/me")
    # == f.http.get ("/v1/accounts/me")
    # f.axios.get ("/v1/accounts/me")
    # f.siesta.v1.accounts ("me").get ()

    def __init__ (self, endpoint, api_call = False):
        self.endpoint = endpoint
        self.s = requests.Session ()
        self._api_call = api_call
        self.siesta = siesta.API (endpoint, reraise_http_error = False)
        self.http2 = HTTP2 (endpoint)
        self.http3 = has_http3 and HTTP3 (endpoint) or None
        if not self._api_call:
            self.axios = Target (endpoint, True)
        else:
            self.set_default_header ('Accept', "application/json")
            self.set_default_header ('Content-Type', "application/json")

    @property
    def http (self):
        return self

    def websocket (self, uri):
        from websocket import create_connection

        u = urlparse (self.endpoint)
        return create_connection ("ws://" + u.netloc + uri)

    def set_default_header (self, k, v):
        self.s.headers.update ({k: v})

    def api (self, point = None):
        if point:
            return siesta.API (point, reraise_http_error = False)
        return self.siesta

    def __enter__ (self):
        return self

    def __exit__ (self, type, value, tb):
        self._close ()

    def __del__ (self):
        self._close ()

    def _close (self):
        pass

    def resolve (self, url):
        if url.startswith ("http://") or url.startswith ("https://"):
            return url
        else:
            return self.endpoint + url

    def _request (self, method, url, *args, **kargs):
        url = self.resolve (url)
        rf = getattr (self.s, method)
        if args:
            args = list (args)
            request_data = args.pop (0)
            args = tuple (args)
        else:
            try:
                request_data = kargs.pop ('data')
            except KeyError:
                request_data = None

        if isinstance (request_data, dict) and self._api_call:
            request_data = json.dumps (request_data)

        if request_data:
            resp = rf (url, request_data, *args, **kargs)
        else:
            resp = rf (url, *args, **kargs)

        if resp.headers.get ('content-type', '').startswith ('application/json'):
            try:
                resp.data = resp.json ()
            except:
                resp.data = {}

            if "__spec__" in resp.data:
                reqh = kargs.get ('headers', {})
                reqh.update (self.s.headers)
                apidoc.log_spec (method.upper (), url, resp.status_code, resp.reason, reqh, request_data, resp.headers, resp.data)
        else:
            resp.data = resp.content
        return resp

    def get (self, url, *args, **karg):
        return self._request ('get', url, *args, **karg)

    def post (self, url, *args, **karg):
        return self._request ('post', url, *args, **karg)

    def upload (self, url, data, **karg):
        files = {}
        for k in list (data.keys ()):
            if isinstance (data [k], IOBase):
                files [k] = data.pop (k)
        return self._request ('post', url, files = files, data = data, **karg)

    def put (self, url, *args, **karg):
        return self._request ('put', url, *args, **karg)

    def patch (self, url, *args, **karg):
        return self._request ('patch', url, *args, **karg)

    def delete (self, url, *args, **karg):
        return self._request ('delete', url, *args, **karg)

    def head (self, url, *args, **karg):
        return self._request ('head', url, *args, **karg)

    def options (self, url, *args, **karg):
        return self._request ('options', url, *args, **karg)

    def rpc (self, url, proxy_class = None):
        return (proxy_class or xmlrpc.client.ServerProxy) (self.resolve (url))
    xmlrpc = rpc

    def jsonrpc (self, url, proxy_class = None):
        import jsonrpclib
        return (proxy_class or jsonrpclib.ServerProxy) (self.resolve (url))

    def grpc (self, url, proxy_class = None):
        raise NotImplementedError

if __name__ == "__main__":
    if "--init" in sys.argv:
        apidoc.truncate_log_dir ()
    if "--make" in sys.argv:
        apidoc.build_doc ()
    if "--clean" in sys.argv:
        apidoc.truncate_log_dir (remove_only = True)





#!/usr/bin/env python3

# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import os
from prettyparse import create_parser
import numpy as np
from precise import __version__

usage = '''
    stdin should be a stream of raw int16 audio, written in
    groups of CHUNK_SIZE samples. If no CHUNK_SIZE is given
    it will read until EOF. For every chunk, an inference
    will be given via stdout as a float string, one per line

    :model str
        Keras or TensorFlow model to read from

    ...
'''

def main (listener_class):
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '0'
    stdout = sys.stdout
    sys.stdout = sys.stderr

    parser = create_parser(usage)
    parser.add_argument('-v', '--version', action='version', version=__version__)
    parser.add_argument('chunk_size', type=int, nargs='?', default=-1,
                        help='Number of bytes to read before making a prediction.'
                             'Higher values are less computationally expensive')
    parser.usage = parser.format_usage().strip().replace('usage: ', '') + ' < audio.wav'
    args = parser.parse_args()

    if sys.stdin.isatty():
        parser.error('Please pipe audio via stdin using < audio.wav')

    listener = listener_class (args.model, args.chunk_size)

    try:
        while True:
            conf = listener.update(sys.stdin.buffer)
            stdout.buffer.write((str(conf) + '\n').encode('ascii'))
            stdout.buffer.flush()
    except (EOFError, KeyboardInterrupt):
        pass


if __name__ == '__main__':
    main()



# ------------------------------------------------
#    Python Siesta
# Siesta is a REST client for python
#
#    Copyright (c) 2008 Rafael Xavier de Souza
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------
#
# Modified at Sep 20, 2017 by Hans Roh
#
# ------------------------------------------------

__version__ = "0.5.2"
__author__ = "Sebastian Castillo <castillobuiles@gmail.com>"
__contributors__ = []

import re
import time
import logging
import json
import requests
import base64
from urllib.parse import urlparse, urlencode
from .attrdict import AttrDict
import html2text
import io
import os
import pprint
from io import IOBase
from . import pathtool
import shutil
import pickle
from .apidoc import log_spec

USER_AGENT = "Python-siesta/%s" % __version__

def typed_urlencode (kw):
    tkw = {}
    for k, v in kw.items ():
        if k [1:3] == "__":
            k = k.replace ("__", ":", 1)
        tkw [k] = v
    return urlencode (tkw)

class Auth(object):
    def encode_params(self, base_url, method, params):
        raise NotImplementedError()

    def makeheaders(self):
        pass


class APIKeyAuth(Auth):
    def __init__(self, api_key, auth_header_name="Authorization"):
        self.api_key = api_key
        self.auth_header_name = auth_header_name

    def encode_params(self):
        pass

    def makeheaders(self):
        return {self.auth_header_name: self.api_key, }


class BasicAuth(Auth):
    def __init__(self, api_username, api_password, auth_header_name="Authorization"):
        self.api_username = api_username
        self.api_password = api_password
        self.auth_header_name = auth_header_name

    def encode_params(self):
        basic_token = base64.encodestring('' + str(self.api_username) + ':' + str(self.api_password))
        basic_token = basic_token.replace('\n', '')
        return basic_token

    def makeheaders(self):
        token = self.encode_params()
        return {self.auth_header_name: 'Basic ' + token, }

class Response (object):
    def __init__ (self, resource, response, method, url, data, headers):
        self.__response = response
        self.resource = resource
        self.data = AttrDict ()

        self.method = method
        self.url = url
        self.request_data = data
        self.request_headers = headers

        self.set_data (response)

    def __getattr__ (self, attr):
        return getattr (self.__response, attr)

    def __str__ (self):
        if not isinstance (self.data, str):
            s = io.StringIO()
            pprint.pprint(self.data, s)
            d = s.getvalue()
        else:
            d = self.data
        return "%d %s\n%s" % (self.__response.status_code, self.__response.reason, d)

    def set_data (self, resp):
        if not resp.text.strip ():
            self.data = None

        else:
            ct = resp.headers.get ('content-type', '')
            if not ct or ct.find ('text/html') == 0:
                h = html2text.HTML2Text()
                h.ignore_links = True
                text = h.handle(resp.text)
                self.data = text
            elif not ct or ct.find ('text/') == 0:
                self.data = resp.text.strip ()
            elif ct.find ('application/json') == 0:
                data = resp.json ()
                if isinstance (data, dict):
                    self.data.update (data)
                    if "__spec__" in self.data:
                        log_spec (self.method.upper (), self.url, self.__response.status_code, self.__response.reason, self.request_headers, self.request_data, self.__response.headers, self.data)
                else:
                    self.data = data
            else:
                self.data = resp.content

        if self.resource._reraise_http_error and not str(resp.status_code).startswith("2"):
            raise AssertionError (self)


class Resource (object):
    ACCEPT = 'application/json'
    CONTENT_TYPE = 'application/json'

    def __init__(self, uri, api, logger, reraise_http_error = True, callback = None):
        self._api = api
        self._uri = uri
        self._logger = logger
        self._reraise_http_error = reraise_http_error
        self._callback = callback
        self._scheme, self._host, self._url = urlparse (self._api.base_url + self._uri) [:3]
        self.headers = {'User-Agent': USER_AGENT, "Accept": self.ACCEPT}

    def __getattr__(self, name):
        key = self._uri + '/' + name
        return self.__class__ (uri=key, api=self._api, logger = self._logger, reraise_http_error = self._reraise_http_error, callback = self._callback)

    def __call__(self, *ids):
        if not ids:
            return self
        key = self._uri
        for id in ids:
            if key:
                key += '/' + str (id)
            else:
                key += str (id)
        return Resource(uri=key, api=self._api, logger = self._logger, reraise_http_error = self._reraise_http_error, callback = self._callback)

    def _request (self, method, data, headers, slash, **kwargs):
        if slash:
            self._url = self._url + "/"
        url = self._url
        if len(kwargs) > 0:
            url = "%s?%s" % (url, typed_urlencode (kwargs))
        headers = headers or {}
        if "Content-Type" not in headers:
            headers ["Content-Type"] = self.CONTENT_TYPE
        if self._api.auth:
            headers.update(self._api.auth.makeheaders())
        if not 'User-Agent' in headers:
            headers['User-Agent'] = self.headers['User-Agent']
        if not 'Accept' in headers and 'Accept' in self.headers:
            headers['Accept'] = self.headers['Accept']

        if self._callback:
            return self._callback (method, url, data, headers)
        return self._continue_request (method, url, data, headers)

    def get(self, headers = None, **kwargs):
        return self._request ('GET', None, headers, False, **kwargs)

    def delete(self, headers = None, **kwargs):
        return self._request ('DELETE', None, headers, False, **kwargs)

    def options(self, headers = None, **kwargs):
        return self._request ('OPTIONS', None, headers, False, **kwargs)

    def post(self, data = None, headers = None, **kwargs):
        return self._request ('POST', data, headers, False, **kwargs)
    upload = post

    def put(self, data = None, headers = None, **kwargs):
        return self._request ('PUT', data, headers, False, **kwargs)

    # For slash ended endpoint
    def patch (self, data = None, headers = None, **kwargs):
        return self._request ('PATCH', data, headers, False, **kwargs)

    def get_ (self, headers = None, **kwargs):
        return self._request ('GET', None, headers, True, **kwargs)

    def delete_ (self, headers = None, **kwargs):
        return self._request ('DELETE', None, headers, True, **kwargs)

    def options_ (self, headers = None, **kwargs):
        return self._request ('OPTIONS', None, headers, True, **kwargs)

    def post_ (self, data, headers = None, **kwargs):
        return self._request ('POST', data, headers, True, **kwargs)

    def put_(self, data, headers = None, **kwargs):
        return self._request ('PUT', data, headers, True, **kwargs)

    def patch_(self, data, headers = None, **kwargs):
        return self._request ('PATCH', data, headers, True, **kwargs)

    def _continue_request(self, method, url, data, headers):
        url = "%s://%s%s" % (self._scheme, self._host, url)
        files = {}
        if isinstance (data, dict):
            for k in list (data.keys ()):
                if isinstance (data [k], IOBase):
                    files [k] = data.pop (k)

        func = getattr (self._api.session, method.lower ())
        if files:
            del headers ["Content-Type"]
            resp = func (url, data = data, files = files, headers = headers, timeout = self._api.REQ_TIMEOUT, verify = False)
        elif data:
            if isinstance (data, dict):
                data = json.dumps (data)
            resp = func (url, data, headers = headers, timeout = self._api.REQ_TIMEOUT, verify = False)
        else:
            resp = func (url, headers = headers, timeout = self._api.REQ_TIMEOUT, verify = False)

        return self._getresponse (resp, method, url, data, headers)

    def _getresponse(self, resp, method, url, data, headers):
        """
        if resp.status_code == 202:
            status_url = resp.getheader('content-location')
            if not status_url:
                raise Exception('Empty content-location from server')

            status_uri = urlparse(status_url).path
            resource = Resource(uri=status_uri, api=self._api, logger = self._logger, reraise_http_error = self._reraise_http_error, callback = self._callback).get()
            retries = 0
            MAX_RETRIES = 3
            resp_status = resource.response.status_code

            while resp_status != 303 and retries < MAX_RETRIES:
                retries += 1
                new_resp.get()
                time.sleep(5)

            if retries == MAX_RETRIES:
                raise Exception('Max retries limit reached without success')

            location = status.conn.getresponse().getheader('location')
            return Resource(uri=urlparse(location).path, api=self._api, logger = self._logger, reraise_http_error = self._reraise_http_error, callback = self._callback).get()
        """
        return Response (self, resp, method, url, data, headers)


class API(object):
    RESOURCE_CLASS = Resource

    REQ_TIMEOUT = 60
    def __init__(self, base_url, auth=None, logger = None, reraise_http_error = True, callback = None):
        self.base_url = base_url + '/' if not base_url.endswith('/') else base_url
        self.api_path = urlparse(base_url).path
        self.session = requests.Session ()
        self.auth = auth
        self.logger = logger
        self.reraise_http_error = reraise_http_error
        self.callback = callback

    def _set_auth (self, auth = None):
        self.auth = auth

    def _set_jwt (self, token = None):
        if token:
            self.auth = APIKeyAuth ("Bearer " + token)
        else:
            self.auth = None

    def __enter__ (self):
        return self

    def __exit__ (self, type, value, tb):
        pass

    def __getattr__(self, name):
        if name in ('get', 'post', 'put', 'patch', 'delete', 'options'):
            r = self.RESOURCE_CLASS (uri='', api=self, logger = self.logger, reraise_http_error = self.reraise_http_error, callback = self.callback)
            return getattr (r, name)
        return self.RESOURCE_CLASS (uri=name, api=self, logger = self.logger, reraise_http_error = self.reraise_http_error, callback = self.callback)

    def __call__(self, id):
        r = self.RESOURCE_CLASS (uri='', api=self, logger = self.logger, reraise_http_error = self.reraise_http_error, callback = self.callback)
        return r (id)





from .driver import Driver, DEFAULT_USER_AGENT
from selenium import webdriver
import os
import time

class Chrome (Driver):
    def config_proxy (self, host, port):
        return {
            'httpProxy' : "%s:%d" %(host, port),
            'sslProxy' : "%s:%d" %(host, port),
            'noProxy' : None,
            'proxyType' : "MANUAL",
            'class' : "org.openqa.selenium.Proxy",
            'autodetect' : False
        }

    def setopt (self, headless, user_agent, window_size, **opts):
        options = webdriver.ChromeOptions ()
        options.add_argument("--disable-gpu")
        headless and options.add_argument('--headless')
        headless and options.add_argument('--no-sandbox')
        headless and options.add_argument('--disable-dev-shm-usage')
        options.add_argument('--window-size={}x{}'.format (*window_size))

        capabilities = options.to_capabilities ()
        if opts.get ("proxy"):
            host, port = self._parse_host (opts ["proxy"])
            capabilities ['proxy'] = self.config_proxy (host, port)
            capabilities['user-agent'] = user_agent
        return capabilities

    def create (self):
        self.driver = webdriver.Chrome(self.driver_path, desired_capabilities = self.capabilities)


class Firefox (Driver):
    def setopt (self, headless, user_agent, window_size, **opts):
        profile = webdriver.FirefoxProfile()
        if opts.get ("proxy"):
            host, port = self.parse_host (opts ["proxy"])
            profile.set_preference("network.proxy.type", 1)
            profile.set_preference("network.proxy.http", host)
            profile.set_preference("network.proxy.http_port", port)
            profile.set_preference("network.proxy.https", host)
            profile.set_preference("network.proxy.https_port", port)
            profile.set_preference("http.response.timeout", self.TIMEOUT)
            profile.set_preference("dom.max_script_run_time", self.TIMEOUT)
            profile.set_preference("general.useragent.override", user_agent)
            profile.update_preferences()
        return profile

    def create (self):
        self.driver = webdriver.Firefox (self.capabilities, executable_path = self.driver_path)


class IE (Chrome):
    def __init__ (self, headless = False, user_agent = DEFAULT_USER_AGENT, window_size = (1600, 900), **kargs):
        Chrome.__init__ (self, None, headless, user_agent, window_size, **kargs)

    def setopt (self, headless, user_agent, window_size, **opts):
        capabilities = webdriver.DesiredCapabilities.INTERNETEXPLORER.copy()
        if opts.get ("proxy"):
            host, port = self.parse_host (opts ["proxy"])
            capabilities['proxy'] = self.config_proxy (host, port)
        return capabilities

    def create (self):
        self.driver = webdriver.Ie (self.capabilities)




# Copyright (C) 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sample that implements a gRPC client for the Google Assistant API."""

import concurrent.futures
import json
import logging
import os
import os.path
import pathlib2 as pathlib
import sys
import time
import uuid
import multiprocessing
import click
import grpc
import google.auth.transport.grpc
import google.auth.transport.requests
import google.oauth2.credentials
import random
import shutil
from .... import pathtool
from google.assistant.embedded.v1alpha2 import (
    embedded_assistant_pb2,
    embedded_assistant_pb2_grpc
)
from tenacity import retry, stop_after_attempt, retry_if_exception
from .grpc import (
    assistant_helpers,
    audio_helpers,
    browser_helpers,
    device_helpers
)
from . import device_register

ASSISTANT_API_ENDPOINT = 'embeddedassistant.googleapis.com'
END_OF_UTTERANCE = embedded_assistant_pb2.AssistResponse.END_OF_UTTERANCE
DIALOG_FOLLOW_ON = embedded_assistant_pb2.DialogStateOut.DIALOG_FOLLOW_ON
CLOSE_MICROPHONE = embedded_assistant_pb2.DialogStateOut.CLOSE_MICROPHONE
PLAYING = embedded_assistant_pb2.ScreenOutConfig.PLAYING
DEFAULT_GRPC_DEADLINE = 60 * 3 + 5


class SampleAssistant(object):
    """Sample Assistant that supports conversations and device actions.

    Args:
      device_model_id: identifier of the device model.
      device_id: identifier of the registered device instance.
      conversation_stream(ConversationStream): audio stream
        for recording query and playing back assistant answer.
      channel: authorized gRPC channel for connection to the
        Google Assistant API.
      deadline_sec: gRPC deadline in seconds for Google Assistant API call.
      device_handler: callback for device actions.
    """

    def __init__(self, language_code, device_model_id, device_id,
                 conversation_stream, display,
                 channel, deadline_sec, device_handler):
        self.language_code = language_code
        self.device_model_id = device_model_id
        self.device_id = device_id
        self.conversation_stream = conversation_stream
        self.display = display

        # Opaque blob provided in AssistResponse that,
        # when provided in a follow-up AssistRequest,
        # gives the Assistant a context marker within the current state
        # of the multi-Assist()-RPC "conversation".
        # This value, along with MicrophoneMode, supports a more natural
        # "conversation" with the Assistant.
        self.conversation_state = None
        # Force reset of first conversation.
        self.is_new_conversation = True

        # Create Google Assistant API gRPC client.
        self.assistant = embedded_assistant_pb2_grpc.EmbeddedAssistantStub(
            channel
        )
        self.deadline = deadline_sec

        self.device_handler = device_handler

    def __enter__(self):
        return self

    def __exit__(self, etype, e, traceback):
        if e:
            return False
        self.conversation_stream.close()

    def is_grpc_error_unavailable(e):
        is_grpc_error = isinstance(e, grpc.RpcError)
        if is_grpc_error and (e.code() == grpc.StatusCode.UNAVAILABLE):
            logging.error('grpc unavailable error: %s', e)
            return True
        return False

    @retry(reraise=True, stop=stop_after_attempt(3), retry=retry_if_exception(is_grpc_error_unavailable))
    def assist(self):
        """Send a voice request to the Assistant and playback the response.

        Returns: True if conversation should continue.
        """
        continue_conversation = False
        device_actions_futures = []

        self.conversation_stream.start_recording()
        logging.info('Recording audio request.')

        def iter_log_assist_requests():
            for c in self.gen_assist_requests():
                assistant_helpers.log_assist_request_without_audio(c)
                yield c
            logging.debug('Reached end of AssistRequest iteration.')

        question = ''
        supplemental_display_text = ''
        detail = ''
        # This generator yields AssistResponse proto messages
        # received from the gRPC Google Assistant API.
        for resp in self.assistant.Assist(iter_log_assist_requests(), self.deadline):
            assistant_helpers.log_assist_response_without_audio(resp)
            if resp.event_type == END_OF_UTTERANCE:
                logging.info('End of audio request detected.')
                logging.info('Stopping recording.')
                self.conversation_stream.stop_recording()
            if resp.speech_results:
                logging.info('Transcript of user request: "%s".',
                             ' '.join(r.transcript
                                      for r in resp.speech_results))
            if len(resp.audio_out.audio_data) > 0:
                self.conversation_stream.write(resp.audio_out.audio_data)
                #if not self.conversation_stream.playing:
                #    self.conversation_stream.stop_recording()
                #    #self.conversation_stream.start_playback()
                #    logging.info('Playing assistant response.')
            if resp.dialog_state_out.conversation_state:
                conversation_state = resp.dialog_state_out.conversation_state
                logging.debug('Updating conversation state.')
                self.conversation_state = conversation_state
            #if resp.dialog_state_out.volume_percentage != 0:
            #    volume_percentage = resp.dialog_state_out.volume_percentage
            #    logging.info('Setting volume to %s%%', volume_percentage)
            #    self.conversation_stream.volume_percentage = volume_percentage
            #if resp.dialog_state_out.microphone_mode == DIALOG_FOLLOW_ON:
                #continue_conversation = True
                #logging.info('Expecting follow-on query from user.')
            #elif resp.dialog_state_out.microphone_mode == CLOSE_MICROPHONE:
            #    continue_conversation = False
            if resp.device_action.device_request_json:
                device_request = json.loads(
                    resp.device_action.device_request_json
                )
                fs = self.device_handler(device_request)
                if fs:
                    device_actions_futures.extend(fs)

            # if self.display and resp.screen_out.data:
            #    system_browser = browser_helpers.system_browser
            #    system_browser.display(resp.screen_out.data)

            if resp.speech_results:
                question = ' '.join(r.transcript for r in resp.speech_results)
            if resp.dialog_state_out.supplemental_display_text and not supplemental_display_text:
                supplemental_display_text = resp.dialog_state_out.supplemental_display_text
            if resp.screen_out.data and not detail:
                detail = resp.screen_out.data

        if len(device_actions_futures):
            logging.info('Waiting for device executions to complete.')
            concurrent.futures.wait(device_actions_futures)

        logging.info('Finished playing assistant response.')
        #self.conversation_stream.stop_playback()
        return question, supplemental_display_text, detail

    def gen_assist_requests(self):
        """Yields: AssistRequest messages to send to the API."""

        config = embedded_assistant_pb2.AssistConfig(
            audio_in_config=embedded_assistant_pb2.AudioInConfig(
                encoding='LINEAR16',
                sample_rate_hertz=self.conversation_stream.sample_rate,
            ),
            audio_out_config=embedded_assistant_pb2.AudioOutConfig(
                encoding='LINEAR16',
                sample_rate_hertz=self.conversation_stream.sample_rate,
                volume_percentage=self.conversation_stream.volume_percentage,
            ),
            dialog_state_in=embedded_assistant_pb2.DialogStateIn(
                language_code=self.language_code,
                conversation_state=self.conversation_state,
                is_new_conversation=self.is_new_conversation,
            ),
            device_config=embedded_assistant_pb2.DeviceConfig(
                device_id=self.device_id,
                device_model_id=self.device_model_id,
            )
        )
        if self.display:
            config.screen_out_config.screen_mode = PLAYING
        # Continue current conversation with later requests.
        self.is_new_conversation = False
        # The first AssistRequest must contain the AssistConfig
        # and no audio data.
        yield embedded_assistant_pb2.AssistRequest(config=config)
        for data in self.conversation_stream:
            # Subsequent requests need audio data, but not config.
            yield embedded_assistant_pb2.AssistRequest(audio_in=data)

logging.basicConfig(level=logging.INFO)
api_endpoint = ASSISTANT_API_ENDPOINT
device_handler = None
tmpdir = '/tmp/rs4/google-assistant'
pathtool.mkdir (tmpdir)

def set_device (project_id_, device_model_id_):
    global api_endpoint, device_model_id, device_id, grpc_channel, device_handler

    project_id = project_id_
    # Load OAuth 2.0 credentials.
    try:
        with open(os.path.join(click.get_app_dir('google-oauthlib-tool'), 'credentials.json'), 'r') as f:
            credentials = google.oauth2.credentials.Credentials(token=None, **json.load(f))
    except Exception as e:
        raise
    device_id, device_model_id = device_register.register (api_endpoint, project_id, device_model_id_)
    http_request = google.auth.transport.requests.Request()
    credentials.refresh(http_request)
    grpc_channel = google.auth.transport.grpc.secure_authorized_channel (credentials, http_request, api_endpoint)
    device_handler = device_helpers.DeviceRequestHandler (device_id)

audio_sample_rate = audio_helpers.DEFAULT_AUDIO_SAMPLE_RATE
audio_sample_width = audio_helpers.DEFAULT_AUDIO_SAMPLE_WIDTH
audio_iter_size = audio_helpers.DEFAULT_AUDIO_ITER_SIZE
audio_block_size = audio_helpers.DEFAULT_AUDIO_DEVICE_BLOCK_SIZE
audio_flush_size = audio_helpers.DEFAULT_AUDIO_DEVICE_FLUSH_SIZE
lock = multiprocessing.Lock ()

def send (audio_source, lang = 'ko-KR', output_audio_file = None, once = True, display = True, verbose = False, *args, **kwargs):
    global grpc_channel, device_model_id, device_id, lock, device_handler

    output_audio_file =  output_audio_file or os.path.join (tmpdir, 'out-' + str (random.random ()) [2:] + '.wav')
    audio_sink = audio_helpers.WaveSink(
        open(output_audio_file, 'wb'),
        sample_rate=audio_sample_rate,
        sample_width=audio_sample_width
    )

    # Create conversation stream with the given audio source and sink.
    conversation_stream = audio_helpers.ConversationStream (
        source = audio_source,
        sink = audio_sink,
        iter_size = audio_iter_size,
        sample_width = audio_sample_width,
    )

    with lock:
        # ensure single thread
        with SampleAssistant (lang, device_model_id, device_id,
                            conversation_stream, display,
                            grpc_channel, DEFAULT_GRPC_DEADLINE,
                            device_handler) as assistant:

            question, supplemental_display_text, detail = assistant.assist ()
            return AssistResponse (question, supplemental_display_text, detail, output_audio_file)

def send_wav (input_audio_file, lang = 'ko-KR', output_audio_file = None, once = True, display = True, verbose = False, *args, **kwargs):
    audio_source = audio_helpers.WaveSource(
        open(input_audio_file, 'rb'),
        sample_rate=audio_sample_rate,
        sample_width=audio_sample_width
    )
    return send (audio_source, lang, output_audio_file, once, display, verbose, *args, **kwargs)
from_audio = send_wav

def use_device (device, lang = 'ko-KR', output_audio_file = None, once = True, display = True, verbose = False, *args, **kwargs):
    audio_source = audio_helpers.SoundDeviceStream(
        sample_rate=audio_sample_rate,
        sample_width=audio_sample_width,
        block_size=audio_block_size,
        flush_size=audio_flush_size,
        device = device
    )
    resp = send (audio_source, lang, output_audio_file, once, display, verbose, *args, **kwargs)
    audio_source.close ()
    return resp

class AssistResponse:
    def __init__ (self, question, supplemental_display_text, detail, output_audio_file):
        self.question = question
        self.supplemental_display_text = supplemental_display_text
        self.detail = detail
        self.audio = output_audio_file

    def move (self, target):
        shutil.move (self.audio, target)
    save = move

    def remove (self):
        if os.path.isfile (self.audio):
            os.remove (self.audio)

if __name__ == "__main__":
    project_id = ''
    device_model_id = ''
    set_device (project_id, device_model_id)
    use_device (None)




#!/usr/bin/env python

import collections
import pyaudio
from . import snowboydetect
import time
import wave
import os
import logging
from ctypes import *
from contextlib import contextmanager

logging.basicConfig()
logger = logging.getLogger("snowboy")
logger.setLevel(logging.INFO)
TOP_DIR = os.path.dirname(os.path.abspath(__file__))

RESOURCE_FILE = os.path.join(TOP_DIR, "resources/common.res")
DETECT_DING = os.path.join(TOP_DIR, "resources/ding.wav")
DETECT_DONG = os.path.join(TOP_DIR, "resources/dong.wav")

def py_error_handler(filename, line, function, err, fmt):
    pass

ERROR_HANDLER_FUNC = CFUNCTYPE(None, c_char_p, c_int, c_char_p, c_int, c_char_p)

c_error_handler = ERROR_HANDLER_FUNC(py_error_handler)

@contextmanager
def no_alsa_error():
    try:
        asound = cdll.LoadLibrary('libasound.so')
        asound.snd_lib_error_set_handler(c_error_handler)
        yield
        asound.snd_lib_error_set_handler(None)
    except:
        yield
        pass

class RingBuffer(object):
    """Ring buffer to hold audio from PortAudio"""

    def __init__(self, size=4096):
        self._buf = collections.deque(maxlen=size)

    def extend(self, data):
        """Adds data to the end of buffer"""
        self._buf.extend(data)

    def get(self):
        """Retrieves data from the beginning of buffer and clears it"""
        tmp = bytes(bytearray(self._buf))
        self._buf.clear()
        return tmp


def play_audio_file(fname=DETECT_DING):
    """Simple callback function to play a wave file. By default it plays
    a Ding sound.
    :param str fname: wave file name
    :return: None
    """
    ding_wav = wave.open(fname, 'rb')
    ding_data = ding_wav.readframes(ding_wav.getnframes())
    with no_alsa_error():
        audio = pyaudio.PyAudio()
    stream_out = audio.open(
        format=audio.get_format_from_width(ding_wav.getsampwidth()),
        channels=ding_wav.getnchannels(),
        rate=ding_wav.getframerate(), input=False, output=True)
    stream_out.start_stream()
    stream_out.write(ding_data)
    time.sleep(0.2)
    stream_out.stop_stream()
    stream_out.close()
    audio.terminate()


class HotwordDetector(object):
    """
    Snowboy decoder to detect whether a keyword specified by `decoder_model`
    exists in a microphone input stream.
    :param decoder_model: decoder model file path, a string or a list of strings
    :param resource: resource file path.
    :param sensitivity: decoder sensitivity, a float of a list of floats.
                              The bigger the value, the more senstive the
                              decoder. If an empty list is provided, then the
                              default sensitivity in the model will be used.
    :param audio_gain: multiply input volume by this factor.
    :param apply_frontend: applies the frontend processing algorithm if True.
    """

    def __init__(self, decoder_model,
                 resource=RESOURCE_FILE,
                 sensitivity=[],
                 audio_gain=1,
                 apply_frontend=False):

        tm = type(decoder_model)
        ts = type(sensitivity)
        if tm is not list:
            decoder_model = [decoder_model]
        if ts is not list:
            sensitivity = [sensitivity]
        model_str = ",".join(decoder_model)

        self.detector = snowboydetect.SnowboyDetect(
            resource_filename=resource.encode(), model_str=model_str.encode())
        self.detector.SetAudioGain(audio_gain)
        self.detector.ApplyFrontend(apply_frontend)
        self.num_hotwords = self.detector.NumHotwords()

        if len(decoder_model) > 1 and len(sensitivity) == 1:
            sensitivity = sensitivity * self.num_hotwords
        if len(sensitivity) != 0:
            assert self.num_hotwords == len(sensitivity), \
                "number of hotwords in decoder_model (%d) and sensitivity " \
                "(%d) does not match" % (self.num_hotwords, len(sensitivity))
        sensitivity_str = ",".join([str(t) for t in sensitivity])
        if len(sensitivity) != 0:
            self.detector.SetSensitivity(sensitivity_str.encode())

        self.ring_buffer = RingBuffer(
            self.detector.NumChannels() * self.detector.SampleRate() * 5)

    def start(self, detected_callback=play_audio_file,
              interrupt_check=lambda: False,
              sleep_time=0.03,
              audio_recorder_callback=None,
              silent_count_threshold=15,
              recording_timeout=100):
        """
        Start the voice detector. For every `sleep_time` second it checks the
        audio buffer for triggering keywords. If detected, then call
        corresponding function in `detected_callback`, which can be a single
        function (single model) or a list of callback functions (multiple
        models). Every loop it also calls `interrupt_check` -- if it returns
        True, then breaks from the loop and return.
        :param detected_callback: a function or list of functions. The number of
                                  items must match the number of models in
                                  `decoder_model`.
        :param interrupt_check: a function that returns True if the main loop
                                needs to stop.
        :param float sleep_time: how much time in second every loop waits.
        :param audio_recorder_callback: if specified, this will be called after
                                        a keyword has been spoken and after the
                                        phrase immediately after the keyword has
                                        been recorded. The function will be
                                        passed the name of the file where the
                                        phrase was recorded.
        :param silent_count_threshold: indicates how long silence must be heard
                                       to mark the end of a phrase that is
                                       being recorded.
        :param recording_timeout: limits the maximum length of a recording.
        :return: None
        """
        self._running = True

        def audio_callback(in_data, frame_count, time_info, status):
            self.ring_buffer.extend(in_data)
            play_data = chr(0) * len(in_data)
            return play_data, pyaudio.paContinue

        with no_alsa_error():
            self.audio = pyaudio.PyAudio()
        self.stream_in = self.audio.open(
            input=True, output=False,
            format=self.audio.get_format_from_width(
                self.detector.BitsPerSample() / 8),
            channels=self.detector.NumChannels(),
            rate=self.detector.SampleRate(),
            frames_per_buffer=2048,
            stream_callback=audio_callback)

        if interrupt_check():
            logger.debug("detect voice return")
            return

        tc = type(detected_callback)
        if tc is not list:
            detected_callback = [detected_callback]
        if len(detected_callback) == 1 and self.num_hotwords > 1:
            detected_callback *= self.num_hotwords

        assert self.num_hotwords == len(detected_callback), \
            "Error: hotwords in your models (%d) do not match the number of " \
            "callbacks (%d)" % (self.num_hotwords, len(detected_callback))

        logger.debug("detecting...")

        state = "PASSIVE"
        while self._running is True:
            if interrupt_check():
                logger.debug("detect voice break")
                break
            data = self.ring_buffer.get()
            if len(data) == 0:
                time.sleep(sleep_time)
                continue

            status = self.detector.RunDetection(data)
            if status == -1:
                logger.warning("Error initializing streams or reading audio data")

            #small state machine to handle recording of phrase after keyword
            if state == "PASSIVE":
                if status > 0: #key word found
                    self.recordedData = []
                    self.recordedData.append(data)
                    silentCount = 0
                    recordingCount = 0
                    message = "Keyword " + str(status) + " detected at time: "
                    message += time.strftime("%Y-%m-%d %H:%M:%S",
                                         time.localtime(time.time()))
                    logger.info(message)
                    callback = detected_callback[status-1]
                    if callback is not None:
                        callback()

                    if audio_recorder_callback is not None:
                        state = "ACTIVE"
                    continue

            elif state == "ACTIVE":
                stopRecording = False
                if recordingCount > recording_timeout:
                    stopRecording = True
                elif status == -2: #silence found
                    if silentCount > silent_count_threshold:
                        stopRecording = True
                    else:
                        silentCount = silentCount + 1
                elif status == 0: #voice found
                    silentCount = 0

                if stopRecording == True:
                    fname = self.saveMessage()
                    audio_recorder_callback(fname)
                    state = "PASSIVE"
                    continue

                recordingCount = recordingCount + 1
                self.recordedData.append(data)

        logger.debug("finished.")

    def saveMessage(self):
        """
        Save the message stored in self.recordedData to a timestamped file.
        """
        filename = 'output' + str(int(time.time())) + '.wav'
        data = b''.join(self.recordedData)

        #use wave to save data
        wf = wave.open(filename, 'wb')
        wf.setnchannels(1)
        wf.setsampwidth(self.audio.get_sample_size(
            self.audio.get_format_from_width(
                self.detector.BitsPerSample() / 8)))
        wf.setframerate(self.detector.SampleRate())
        wf.writeframes(data)
        wf.close()
        logger.debug("finished saving: " + filename)
        return filename

    def terminate(self):
        """
        Terminate audio stream. Users can call start() again to detect.
        :return: None
        """
        self.stream_in.stop_stream()
        self.stream_in.close()
        self.audio.terminate()
        self._running = False


import time
from datetime import datetime

class Timer:
	def __init__ (self):		
		self.jobs = {}
		self.lastest = ''
		self.created = time.time ()
		self._begun = False
		
	def begin (self, name = 'job'):
		assert not self._begun
		self._begun = True
		self.s = time.time ()
		self.jobs [name] = time.time ()
		self.lastest = name
		print ('* Started {}'.format (name))
		
	def finish (self, tail = '', name = ''):
		assert self._begun
		self._begun = False
		jobname = name or self.lastest
		due = time.time () - self.jobs [jobname]
		print ('* Finished {} for {:2.4f} seconds{}'.format (jobname, due, tail and ', ' + tail or ''))
		self.s = time.time ()
		self.lastest = ''
		
	
	def close (self):
		due = time.time () - self.created
		print ('-------------------')
		print ('* Finished script for {:2.4f} seconds'.format (due))


class Timer2:
	def __init__ (self):
		self.stime = 0
		
	def begin (self):
		self.stime = datetime.now ()
	
	def finish (self, n = 0):
		diff = datetime.now () - self.stime
		print ('Duration: {}'.format (diff))
		if n:
			sec = diff.seconds + (diff.microseconds / 1000000)
			print ('Speed: {:.3f} IPS'.format (n / sec))
		    
if __name__ == "__main__":
    detect ('face560x840.jpg', 100)
    detect ('face250x250.jpg', 100)
    

		